#!/usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2019 CKIP Lab'
__license__ = 'CC BY-NC-SA 4.0'

import datetime as _datetime
import os as _os
import sys as _sys
import warnings as _warnings

import ckipnlp as _about

def create_ws_ini(*, # pylint: disable=too-many-statements
    data2dir=None,
    lexfile=None,
    new_style_format=False,
    show_category=True,
    sentence_max_word_num=80,
    **options,
):
    """Generate CKIP word segmentation config.

    Parameters
    ----------
        data2dir : str
            the path to the folder "Data2/".
        lexfile : str
            the path to the user-defined lexicon file.

        new_style_format : bool
            split sentences by newline characters ("\\\\n") rather than punctuations.
        show_category : bool
            show part-of-speech tags.

        sentence_max_word_num : int
            maximum number of words per sentence.
    """

    # pylint: disable=invalid-name

    if data2dir is None:
        data2dir = _os.getenv('CKIPWS_DATA2')
        if not data2dir:
            data2dir = _os.path.join(_sys.prefix, 'share', 'ckipnlp', 'Data2')
        if not _os.path.isdir(data2dir):
            _warnings.warn('Invalid data2dir (%s)' % data2dir)
            data2dir = 'Data2'

    cfg = []

    cfg.append(';PyCkip {version}'.format(version=_about.__version__))
    cfg.append(';ws.ini')
    cfg.append(';Auto-generated {date}'.format(date=_datetime.datetime.now()))
    cfg.append('')

    cfg.append('[ConsoleLogger]')
    cfg.append('Name=ConsoleLogger')
    cfg.append('')

    if lexfile:
        cfg.append('[CTextLexicon]')
        cfg.append('Name=TextLex')
        cfg.append('FileName={lexfile}'.format(lexfile=lexfile))
        cfg.append('')

    cfg.append('[CLexicon]')
    cfg.append('Name=Lex')
    cfg.append('FileName={data2dir}/Lexicon.Dat'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[CALexicon]')
    cfg.append('Name=CALex')
    cfg.append('FileName={data2dir}/CALexicon.Dat'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[CDMMergedParser]')
    cfg.append('Name=DMMergedParser')
    cfg.append('GenerateMaxLengthWordOnly=no')
    cfg.append('LexiconName=Lex')
    cfg.append('')

    cfg.append('[CDMSplittedParser]')
    cfg.append('Name=DMSplittedParser')
    cfg.append('GenerateMaxLengthWordOnly=no')
    cfg.append('LexiconName=Lex')
    cfg.append('')

    cfg.append('[CHTRDRule3]')
    cfg.append('Name=RD3')
    cfg.append('LexiconName=Lex')
    cfg.append('')

    cfg.append('[CHTRDRule6]')
    cfg.append('Name=RD6')
    cfg.append('LexiconName=Lex')
    cfg.append('')

    cfg.append('[CHTRDRule7]')
    cfg.append('Name=RD7')
    cfg.append('LexiconName=Lex')
    cfg.append('')

    cfg.append('[CHTForeignWord]')
    cfg.append('Name=FW')
    cfg.append('LexiconName=Lex')
    cfg.append('')

    cfg.append('[CHTBoundWord]')
    cfg.append('Name=BW')
    cfg.append('LexiconName=Lex')
    cfg.append('')

    cfg.append('[CMaxMatch]')
    cfg.append('Name=MaxMatch')
    cfg.append('WindowSize=3')
    cfg.append('')

    cfg.append('[CHTCategoryPredictor]')
    cfg.append('Name=CatPred')
    cfg.append('PrefixCategoryFileName={data2dir}/CatPredictData/PrefixCategoryFreq'.format(data2dir=data2dir))
    cfg.append('PrefixFileName={data2dir}/CatPredictData/PrefixFreq'.format(data2dir=data2dir))
    cfg.append('SuffixCategoryFileName={data2dir}/CatPredictData/SuffixCategoryFreq'.format(data2dir=data2dir))
    cfg.append('SuffixFileName={data2dir}/CatPredictData/SuffixFreq'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[CStatProb]')
    cfg.append('Name=CAProb1')
    cfg.append('FileName={data2dir}/CAStat-w(0)c(0)-w(-1).dat'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[CStatProb]')
    cfg.append('Name=CAProb2')
    cfg.append('FileName={data2dir}/CAStat-w(0)c(0)-w(1).dat'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[CStatProb]')
    cfg.append('Name=CAProb3')
    cfg.append('FileName={data2dir}/CAStat-w(0)c(0)-w(-2).dat'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[CStatProb]')
    cfg.append('Name=CAProb4')
    cfg.append('FileName={data2dir}/CAStat-w(0)c(0)-w(2).dat'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[CoveringAmbiguity]')
    cfg.append('Name=CA')
    cfg.append('LexiconName=Lex')
    cfg.append('CoveringAmbiguityLexiconName=CALex')
    cfg.append('InsertSplittedWordsOnly=false')
    cfg.append('StatisticProbability1=CAProb1')
    cfg.append('StatisticProbability2=CAProb2')
    cfg.append('StatisticProbability3=CAProb3')
    cfg.append('StatisticProbability4=CAProb4')
    cfg.append('')

    cfg.append('[CStatProb]')
    cfg.append('Name=Prob1')
    cfg.append('FileName={data2dir}/CKIPWStatistic-w(-1)-w(0).dat'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[CStatProb]')
    cfg.append('Name=Prob2')
    cfg.append('FileName={data2dir}/CKIPWStatistic-c(-1)-c(0).dat'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[CStatProb]')
    cfg.append('Name=Prob3')
    cfg.append('FileName={data2dir}/CKIPWStatistic-c(0)-w(0).dat'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[CSimpleProbModel]')
    cfg.append('Name=ProbModel')
    cfg.append('StatisticProbability1=Prob1')
    cfg.append('StatisticProbability2=Prob2')
    cfg.append('StatisticProbability3=Prob3')
    cfg.append('LexiconName=Lex')
    if lexfile:
        cfg.append('TextLexiconName=TextLex')
    cfg.append('AdjustProb3=true')
    cfg.append('CoveringAmbiguityLexiconName=CALex')
    cfg.append('CategoryPredictor=CatPred')
    cfg.append('KeepBestCategory=true')
    cfg.append('SimplifiedCategory=false')
    cfg.append('')

    cfg.append('[CDetectMonosyllabicMorpheme]')
    cfg.append('Name=DMM')
    cfg.append('ApplyDefaultHeuristicDetectRule=yes')
    cfg.append('InitDataPath={data2dir}/uwea/qrulepool/'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[ChineseName]')
    cfg.append('Name=CN')
    cfg.append('LexiconName=Lex')
    cfg.append('InitDataPath={data2dir}/uwea/data/'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[CForeignName]')
    cfg.append('Name=FN')
    cfg.append('LexiconName=Lex')
    if lexfile:
        cfg.append('TextLexiconName=TextLex')
    cfg.append('InitDataPath={data2dir}/uwea/data/'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[CompoundWord]')
    cfg.append('Name=CW')
    cfg.append('LexiconName=Lex')
    cfg.append('InitDataPath={data2dir}/uwea/data/'.format(data2dir=data2dir))
    cfg.append('')

    cfg.append('[CStatisticWord]')
    cfg.append('Name=SW')
    cfg.append('LexiconName=Lex')
    if lexfile:
        cfg.append('TextLexiconName=TextLex')
    cfg.append('CategoryPredictor=CatPred')
    cfg.append('InitDataPath={data2dir}/uwea/data/'.format(data2dir=data2dir))
    cfg.append('ApplyRule=639')
    cfg.append('')

    cfg.append('[CAffixCombiner]')
    cfg.append('Name=AC')
    cfg.append('LexiconName=Lex')
    if lexfile:
        cfg.append('TextLexiconName=TextLex')
    cfg.append('CategoryPredictor=CatPred')
    cfg.append('')

    cfg.append('[CSimilarStructureCombiner]')
    cfg.append('Name=SSC')
    cfg.append('AutoCombineWordLen=2')
    cfg.append('HeuristicCombinedWordMaxLen=3')
    cfg.append('LexiconName=Lex')
    cfg.append('CategoryPredictor=CatPred')
    cfg.append('')

    cfg.append('[COnlineLexicon]')
    cfg.append('Name=OnlineLexForUWGen')
    cfg.append('')

    cfg.append('[CUnknownWord]')
    cfg.append('Name=UW')
    cfg.append('UnknownWordGeneratorList=DMM CN FN CW SW SSC')
    cfg.append('OnlineLexicon=OnlineLexForUWGen')
    cfg.append('')

    if lexfile:
        cfg.append('[CLexWordGenerator]')
        cfg.append('Name=myLWGen')
        cfg.append('LexiconName=TextLex')
        cfg.append('')

    cfg.append('[CLexWordGenerator]')
    cfg.append('Name=LWGen')
    cfg.append('LexiconName=Lex')
    cfg.append('')

    cfg.append('[CLexWordGenerator]')
    cfg.append('Name=LWGen1')
    cfg.append('LexiconName=Lex')
    cfg.append('MaxWordLen=1')
    cfg.append('')

    cfg.append('[CLexWordGenerator]')
    cfg.append('Name=UWGen')
    cfg.append('LexiconName=OnlineLexForUWGen')
    cfg.append('')

    cfg.append('[CSimpleProbModelResult]')
    cfg.append('Name=ProbModelResult')
    cfg.append('ProbabilityModelName=ProbModel')
    cfg.append('NewStyleFormat={NewStyleFormat}'.format(NewStyleFormat=str(new_style_format).lower()))
    cfg.append('ShowCategory={ShowCategory}'.format(ShowCategory=str(show_category).lower()))
    cfg.append('LexiconName=Lex')
    cfg.append('CategoryPredictor=CatPred')
    cfg.append('KeepExistingWord=true')
    cfg.append('FeatureAssigner=FA')
    cfg.append('FilterBadWord=false')
    cfg.append('')

    cfg.append('[CDetectDMForPostProcess]')
    cfg.append('Name=DDMFPP')
    cfg.append('')

    cfg.append('[CRemoveWordToBePostProcessed]')
    cfg.append('Name=RWTBPP')
    cfg.append('')

    HandlerList = [
        'LWGen', 'myLWGen', 'DMMergedParser', 'RD3', 'RD6', 'RD7', 'FW', 'BW', 'MaxMatch', 'ProbModel', 'UW', 'DDMFPP',
        'LWGen', 'UWGen', 'RWTBPP', 'LWGen', 'myLWGen', 'DMSplittedParser', 'BW', 'MaxMatch', 'ProbModel', 'CA'
    ]

    if not lexfile:
        while 'myLWGen' in HandlerList:
            HandlerList.remove('myLWGen')

    cfg.append('[CWordSegmentor]')
    cfg.append('Name=MainWS')
    cfg.append('ArticleMaxLineNum=300')
    cfg.append('SentenceMaxWordNum={sentence_max_word_num}'.format(sentence_max_word_num=sentence_max_word_num))
    cfg.append('ReloadMyDic=false')
    cfg.append('SentenceDelimiter=，,；。！？')
    cfg.append('HandlerList={HandlerList}'.format(HandlerList=' '.join(HandlerList)))
    cfg.append('Result=ProbModelResult')
    cfg.append('')

    return '\n'.join(cfg), options

def create_parser_ini(*, # pylint: disable=too-many-statements
    wsinifile,
    ruledir=None,
    rdbdir=None,
    do_ws=True,
    do_parse=True,
    do_role=True,
    sentence_delim='，,；。！？',
    **options,
):
    """Generate CKIP parser config.

    Parameters
    ----------
        ruledir : str
            the path to "Rule/".
        rdbdir : str
            the path to "RDB/".

        do_ws : bool
            do word-segmentation.
        do_parse : bool
            do parsing.
        do_role : bool
            do role.

        sentence_delim : str
            the sentence delimiters.
    """

    # pylint: disable=invalid-name

    if ruledir is None:
        ruledir = _os.getenv('CKIPPARSER_RULE')
        if not ruledir:
            ruledir = _os.path.join(_sys.prefix, 'share', 'ckipnlp', 'Rule')
        if not _os.path.isdir(ruledir):
            _warnings.warn('Invalid ruledir (%s)' % ruledir)
            ruledir = 'Rule'

    if rdbdir is None:
        rdbdir = _os.getenv('CKIPPARSER_RDB')
        if not rdbdir:
            rdbdir = _os.path.join(_sys.prefix, 'share', 'ckipnlp', 'RDB')
        if not _os.path.isdir(rdbdir):
            _warnings.warn('Invalid rdbdir (%s)' % rdbdir)
            rdbdir = 'RDB'

    IsTag = not do_ws
    AssignRole = do_role
    AssignRoleOnly = False

    if not do_parse:
        if not do_ws and not do_role:
            raise ValueError('Must select at least one task')
        if do_ws and not do_role:
            raise ValueError('Use ckipws.CkipWs for word-segmentation')
        if not do_ws and do_role:
            AssignRoleOnly = True
        if do_ws and do_role:
            raise ValueError('Invalid tasks')

    cfg = []

    cfg.append(';PyCkip {version}'.format(version=_about.__version__))
    cfg.append(';parser.ini')
    cfg.append(';Auto-generated {date}'.format(date=_datetime.datetime.now()))
    cfg.append('')

    cfg.append('[WordSeg]')
    cfg.append('ini={wsinifile}'.format(wsinifile=wsinifile))
    cfg.append('')

    cfg.append('[Parser]')
    cfg.append('SetPos13=0')
    cfg.append('13CateFile={ruledir}/13Cate.txt'.format(ruledir=ruledir))
    cfg.append('')

    cfg.append('SetMap=1')
    cfg.append('CatMapFile={ruledir}/CatMap.txt'.format(ruledir=ruledir))
    cfg.append('WordLib1={ruledir}/WordLib1.txt'.format(ruledir=ruledir))
    cfg.append('WordLib2={ruledir}/WordLib2.txt'.format(ruledir=ruledir))
    cfg.append('WordLib3={ruledir}/WordLib3.txt'.format(ruledir=ruledir))
    cfg.append('')

    cfg.append('GrammarRule={ruledir}/CKIP-Rule.txt'.format(ruledir=ruledir))
    cfg.append('HeadRule={ruledir}/CKIP-Head.txt'.format(ruledir=ruledir))
    cfg.append('')

    cfg.append('SetChangePos=1')
    cfg.append('SentenceDelimiter={SentenceDelimiter}'.format(SentenceDelimiter=sentence_delim))
    cfg.append('SetLength=15')
    cfg.append('NormalPos=1')
    cfg.append('NormalTree=1')
    cfg.append('IsTag={IsTag}'.format(IsTag=int(IsTag)))
    cfg.append('')

    cfg.append('[SRL]')
    cfg.append('DataPath={rdbdir}/'.format(rdbdir=rdbdir))
    cfg.append('AssignRole={AssignRole}'.format(AssignRole=int(AssignRole)))
    cfg.append('AssignRoleOnly={AssignRoleOnly}'.format(AssignRoleOnly=int(AssignRoleOnly)))

    return '\n'.join(cfg), options



# -*- coding:utf-8 -*-
# cython: language_level=3

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2019 CKIP Lab'
__license__ = 'CC BY-NC-SA 4.0'

cimport src.ws.cckipws as cckipws
from libc.stdlib cimport malloc, free
from cpython.unicode cimport PyUnicode_AsUnicode

cdef class CkipWsCore:

    cdef cckipws.wordseg_t __obj

    def __cinit__(self):
        self.__obj = cckipws.WordSeg_New()

    def __dealloc__(self):
        if self.__obj is not NULL:
            cckipws.WordSeg_Destroy(self.__obj)
            pass

    def init_data(self, inifile):
        ret = cckipws.WordSeg_InitData(self.__obj, inifile.encode())
        if not ret:
            raise IOError()

    def enable_logger(self):
        cckipws.WordSeg_EnableConsoleLogger(self.__obj)

    def apply_list(self, ilist):
        inum = len(ilist)

        iarr = <const Py_UNICODE**> malloc(sizeof(const Py_UNICODE*) * inum)
        for i in range(inum):
            iarr[i] = PyUnicode_AsUnicode(ilist[i])
        ret = cckipws.WordSeg_ApplyList(self.__obj, inum, iarr)
        free(iarr)
        assert ret is not None

        olist = []
        result = cckipws.WordSeg_GetResultBegin(self.__obj)
        while result is not NULL:
            olist.append(result.strip())
            result = cckipws.WordSeg_GetResultNext(self.__obj)

        return olist

    def apply_file(self, ifile, ofile, uwfile):
        ret = cckipws.WordSeg_ApplyFile(self.__obj, ifile.encode(), ofile.encode(), uwfile.encode())
        assert ret is not None



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2019 CKIP Lab'
__license__ = 'CC BY-NC-SA 4.0'

import os as _os
import tempfile as _tempfile
import warnings as _warnings

try:
    if not _os.environ.get('_SPHINX'):
        from ckipnlp._core.ws import CkipWsCore
except ImportError as exc:
    raise ImportError(
        'Please reinstall ‘ckipnlp’ with ‘--install-option=\'--ws\'’. '
        'Visit https://pypi.org/project/ckipnlp for more information.'
    ) from exc
except Exception as exc:
    raise exc

from ckipnlp.util.ini import create_ws_ini

class CkipWs:
    """The CKIP word segmentation driver.

    Parameters
    ----------
        logger : bool
            enable logger.
        inifile : str
            the path to the INI file.
    Other Parameters
    ----------------
        **
            the configs for CKIPWS, ignored if **inifile** is set. Please refer :func:`ckipnlp.util.ini.create_ws_ini`.

    Warning
    -------
        Never instance more than one object of this class!
    """

    def __init__(self, *,
        logger=False,
        inifile=None,
        **kwargs,
    ):

        self.__core = CkipWsCore()

        if logger:
            self.__core.enable_logger()

        if not inifile:
            fini = _tempfile.NamedTemporaryFile(mode='w')
            inifile = fini.name
            inidata, kwargs = create_ws_ini(**kwargs)
            fini.write(inidata)
            fini.flush()

        def CkipWs(*, _=None): pass # pylint: disable=redefined-outer-name, invalid-name, multiple-statements
        CkipWs(**kwargs)

        self.__core.init_data(inifile)

        try:
            fini.close()
        except: # pylint: disable=bare-except
            pass

    def __call__(self, text):
        return self.apply(text)

    def apply(self, text):
        """Segment a sentence.

        Parameters
        ----------
            text : str
                the input sentence.

        Return
        ------
            str
                the output sentence.

        .. note::
            One may also call this method as :func:`__call__`.
        """
        return self.apply_list([text])[0]

    def apply_list(self, ilist):
        """Segment a list of sentences.

        Parameters
        ----------
            ilist
                the list of input sentences.

        Return
        ------
            List[str]
                the list of output sentences.
        """
        return self.__core.apply_list(ilist)

    def apply_file(self, ifile, ofile, uwfile=''):
        """Segment a file.

        Parameters
        ----------
            ifile : str
                the input file.
            ofile : str
                the output file (will be overwritten).
            uwfile : str
                the unknown word file (will be overwritten).
        """
        return self.__core.apply_file(ifile, ofile, uwfile)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import warnings
warnings.warn('Module ‘ckipws’ is deprecated. Please import ‘ckipnlp.ws’ instead.', FutureWarning)
from ckipnlp.ws import *
from ckipnlp import __version__

class CkipWS(CkipWs):
    warnings.warn('Class ‘CkipWS’ is deprecated. Please use ‘CkipWs’ instead.', FutureWarning)



# -*- coding:utf-8 -*-

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2019 CKIP Lab'
__license__ = 'CC BY-NC-SA 4.0'

cdef extern:

	ctypedef void* wordseg_t

	wordseg_t WordSeg_New()
	int WordSeg_InitData(wordseg_t obj, char *FileName);
	int WordSeg_ApplyFile(wordseg_t obj, char *input, char *output, char *uwfile);
	int WordSeg_ApplyList(wordseg_t obj, int length, const Py_UNICODE **inputList);
	int WordSeg_ApplyArticle(wordseg_t obj, int length, const Py_UNICODE **inputList);
	const Py_UNICODE* WordSeg_GetResultBegin(wordseg_t obj);
	const Py_UNICODE* WordSeg_GetResultNext(wordseg_t obj);
	const Py_UNICODE* WordSeg_GetUWBegin(wordseg_t obj);
	const Py_UNICODE* WordSeg_GetUWNext(wordseg_t obj);
	void WordSeg_EnableConsoleLogger(wordseg_t obj);
	void WordSeg_Destroy(wordseg_t obj);



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2019 CKIP Lab'
__license__ = 'CC BY-NC-SA 4.0'

import collections as _collections
import json as _json

from typing import (
    NamedTuple,
)

class WsWord(NamedTuple):
    """A word-segmented word."""

    word: str #: *str* – the word.
    pos: str  #: *str* – the post-tag.

    @classmethod
    def from_text(cls, text):
        """Create a :class:`WsWord` object from :class:`ckipnlp.ws.CkipWs` output.

        Parameters
        ----------
            text : str
                A word from :class:`ckipnlp.ws.CkipWs` output.
        """
        return cls(*text.strip(')').rsplit('(', 1))

    def __str__(self):
        return self.to_text()

    def to_text(self):
        """Transform to plain text."""
        return '{}({})'.format(self.word, self.pos)

    def to_dict(self):
        """Transform to python dict/list."""
        return self._asdict() # pylint: disable=no-member

    def to_json(self, **kwargs):
        """Transform to JSON format."""
        return _json.dumps(self.to_dict(), **kwargs)

class WsSentence(_collections.UserList): # pylint: disable=too-many-ancestors
    """A word-segmented sentence."""

    item_class = WsWord

    @classmethod
    def from_text(cls, text):
        """Create :class:`WsSentence` object from :class:`ckipnlp.ws.CkipWs` output.

        Parameters
        ----------
            text : str
                A sentence from :class:`ckipnlp.ws.CkipWs` output.
        """
        return cls(map(cls.item_class.from_text, text.split('\u3000')))

    def __str__(self):
        return self.to_text()

    def to_text(self):
        """Transform to plain text."""
        return '\u3000'.join(map(self.item_class.to_text, self))

    def to_dict(self):
        """Transform to python dict/list."""
        return [word.to_dict() for word in self]

    def to_json(self, **kwargs):
        """Transform to JSON format."""
        return _json.dumps(self.to_dict(), **kwargs)


class WsSentenceList(_collections.UserList): # pylint: disable=too-many-ancestors
    """A list of word-segmented sentence."""

    item_class = WsSentence

    @classmethod
    def from_text(cls, text_list):
        """Create :class:`WsSentenceList` object from :class:`ckipnlp.ws.CkipWs` output.

        Parameters
        ----------
            text_list : List[str]
                A list of sentence from :class:`ckipnlp.ws.CkipWs` output.
        """
        return cls(map(cls.item_class.from_text, text_list))

    def __str__(self):
        return self.to_text()

    def to_text(self):
        """Transform to plain text."""
        return '\n'.join(map(self.item_class.to_text, self))

    def to_dict(self):
        """Transform to python dict/list."""
        return [sentence.to_dict() for sentence in self]

    def to_json(self, **kwargs):
        """Transform to JSON format."""
        return _json.dumps(self.to_dict(), **kwargs)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import warnings
warnings.warn('Module ‘ckipparser’ is deprecated. Please import ‘ckipnlp.parser’ instead.', FutureWarning)
from ckipnlp.parser import *
from ckipnlp import __version__



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2019 CKIP Lab'
__license__ = 'CC BY-NC-SA 4.0'

from setuptools import dist
dist.Distribution().fetch_build_eggs([
    'Cython>=0.29',
])

################################################################################

import os
import sys
import warnings

from distutils.version import StrictVersion
import setuptools

assert StrictVersion(setuptools.__version__) >= StrictVersion('40.0'), \
    'Please update setuptools to 40.0+ using `pip install -U setuptools`.'

################################################################################

from setuptools import setup, find_namespace_packages
from setuptools.extension import Extension
from setuptools.command.install import install
from setuptools.command.develop import develop
from Cython.Build import cythonize

import ckipnlp as about

with open('README.rst', encoding='utf-8') as fin:
    readme = fin.read()

################################################################################

def os_environ_append(name, dirpath):
    if name in os.environ:
        os.environ[name] += os.pathsep + dirpath
    else:
        os.environ[name] = dirpath

def os_environ_prepend(name, dirpath):
    if name in os.environ:
        os.environ[name] = dirpath + os.pathsep + os.environ[name]
    else:
        os.environ[name] = dirpath

class CommandMixin(object):

    user_options = [
        ('ws',        None, 'with CKIPWS [default]'),
        ('parser',    None, 'with CKIP-Parser [default]'),
        ('no-ws',     None, 'without CKIPWS'),
        ('no-parser', None, 'without CKIP-Parser'),

        ('ws-dir=',       None, 'CKIPWS root directory'),
        ('ws-lib-dir=',   None, 'CKIPWS libraries directory [default is <ws-dir>/lib]'),
        ('ws-share-dir=', None, 'CKIPWS share directory [default is <ws-dir>]'),

        ('parser-dir=',       None, 'CKIP-Parser root directory'),
        ('parser-lib-dir=',   None, 'CKIP-Parser libraries directory [default is "<parser-dir>/lib"]'),
        ('parser-share-dir=', None, 'CKIPWS share directory [default is "<parser-dir>"]'),

        ('data2-dir=', None, 'CKIPWS "Data2" directory [default is "<ws-share-dir>/Data2" or "<parser-share-dir>/Data2"]'),
        ('rule-dir=',  None, 'CKIP-Parser "Rule" directory [default is "<parser-share-dir>/Rule"]'),
        ('rdb-dir=',   None, 'CKIP-Parser "RDB" directory [default is "<parser-share-dir>/RDB"]'),
    ]

    negative_opt = {
        'no-ws':     'ws',
        'no-parser': 'parser',
    }

    def initialize_options(self):
        self.ws     = False
        self.parser = False

        self.ws_dir       = None
        self.ws_lib_dir   = None
        self.ws_share_dir = None

        self.parser_dir = None
        self.parser_lib_dir   = None
        self.parser_share_dir = None

        self.data2_dir = None
        self.rule_dir  = None
        self.rdb_dir   = None

        super(CommandMixin, self).initialize_options()

    def finalize_options(self):

        # subdirectory
        opt_subdirectory = [
            ('ws_lib_dir',      'ws_dir',     'lib',),
            ('ws_share_dir',    'ws_dir',     '',),
            ('parser_lib_dir',  'parser_dir', 'lib',),
            ('parser_share_dir','parser_dir', '',),

            ('data2_dir', 'ws_share_dir',     'Data2',),
            ('data2_dir', 'parser_share_dir', 'Data2',),
            ('rule_dir',  'parser_share_dir', 'Rule',),
            ('rdb_dir',   'parser_share_dir', 'RDB',),
        ]
        for opt0, opt1, subdir in opt_subdirectory:
            dir0 = getattr(self, opt0)
            dir1 = getattr(self, opt1)
            if not dir0 and dir1:
                setattr(self, opt0, os.path.join(dir1, subdir))

        # directory
        opt_directory = [
            'ws_lib_dir',
            'parser_lib_dir',
            'data2_dir',
            'rule_dir',
            'rdb_dir',
        ]
        for opt0 in opt_directory:
            dir0 = getattr(self, opt0)
            if dir0 and not os.path.isdir(dir0):
                raise IOError('--%s (%s) is not a directory' % (opt0.replace('_', '-'), dir0,))

        super(CommandMixin, self).finalize_options()

    def run(self):

        # CKIPWS
        if self.ws:
            print('- Enable CKIPWS support')
            if self.ws_lib_dir:
                print('- Use CKIPWS library from (%s)' % self.ws_lib_dir)
                i = next((i for i, em in enumerate(self.distribution.ext_modules) if em.name == 'ckipnlp._core.ws'), None)
                self.distribution.ext_modules[i].library_dirs.append(self.ws_lib_dir)
                self.distribution.ext_modules[i].runtime_library_dirs.append(self.ws_lib_dir)
        else:
            print('- Disable CKIPWS support')
            i = next((i for i, em in enumerate(self.distribution.ext_modules) if em.name == 'ckipnlp._core.ws'), None)
            if i is not None: del self.distribution.ext_modules[i]

        # CKIP-Parser
        if self.parser:
            print('- Enable CKIP-Parser support')
            if self.parser_lib_dir:
                print('- Use CKIP-Parser library from (%s)' % self.parser_lib_dir)
                i = next((i for i, em in enumerate(self.distribution.ext_modules) if em.name == 'ckipnlp._core.parser'), None)
                self.distribution.ext_modules[i].library_dirs.append(self.parser_lib_dir)
                self.distribution.ext_modules[i].runtime_library_dirs.append(self.parser_lib_dir)
        else:
            print('- Disable CKIP-Parser support')
            i = next((i for i, em in enumerate(self.distribution.ext_modules) if em.name == 'ckipnlp._core.parser'), None)
            if i is not None: del self.distribution.ext_modules[i]

        # Data
        if self.data2_dir:
            print('- Use "Data2" from (%s)' % self.data2_dir)
            self.data_files('share/ckipnlp/Data2/', self.data2_dir)

        if self.rule_dir:
            print('- Use "Rule" from (%s)' % self.rule_dir)
            self.data_files('share/ckipnlp/Rule/', self.rule_dir)

        if self.rdb_dir:
            print('- Use "RDB" from (%s)' % self.rdb_dir)
            self.data_files('share/ckipnlp/RDB/', self.rdb_dir)

        # Python packages
        # self.distribution.packages = list(self.distribution.package_dir.keys())

        super(CommandMixin, self).run()

    def data_files(self, prefix, dirtop):
        for dirpath, _, files in os.walk(dirtop):
            self.distribution.data_files.append((
                os.path.join(prefix, os.path.relpath(dirpath, dirtop)),
                [os.path.join(dirpath, file) for file in files],
            ))

class InstallCommand(CommandMixin, install):
    user_options = install.user_options + CommandMixin.user_options
    negative_opt = install.negative_opt
    negative_opt.update(CommandMixin.negative_opt)

    def __init__(self, *args, **kwargs):
        install.__init__(self, *args, **kwargs)

class DevelopCommand(CommandMixin, develop):
    user_options = develop.user_options + CommandMixin.user_options
    negative_opt = develop.negative_opt
    negative_opt.update(CommandMixin.negative_opt)

    def __init__(self, *args, **kwargs):
        develop.__init__(self, *args, **kwargs)

################################################################################

setup(
    name=about.__name__,
    version=about.__version__,
    author=about.__author_name__,
    author_email=about.__author_email__,
    description=about.__description__,
    long_description=readme,
    long_description_content_type='text/x-rst',
    url=about.__url__,
    download_url=about.__download_url__,
    platforms=['linux_x86_64'],
    license=about.__license__,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Cython',
        'License :: Free for non-commercial use',
        'Operating System :: POSIX :: Linux',
        'Natural Language :: Chinese (Traditional)',
    ],
    python_requires='>=3.5',
    packages=find_namespace_packages(include=['ckipnlp', 'ckipnlp.*', 'ckipws', 'ckipparser',]),
    install_requires=[
        'treelib>=1.5.5',
    ],
    ext_modules=cythonize(
        [
            Extension('ckipnlp._core.ws',
                sources=['src/ws/ckipws.pyx'],
                libraries=['WordSeg'],
            ),
            Extension('ckipnlp._core.parser',
                sources=['src/parser/ckipparser.pyx'],
                libraries=['CKIPCoreNLP','CKIPParser','CKIPWS','CKIPSRL'],
            ),
        ],
        build_dir='build',
    ),
    data_files=[],
    cmdclass={
        'install': InstallCommand,
        'develop': DevelopCommand,
    },
)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2019 CKIP Lab'
__license__ = 'CC BY-NC-SA 4.0'

import collections as _collections
import itertools as _itertools
import json as _json

from typing import (
    NamedTuple,
)

import treelib as _treelib

################################################################################################################################

class ParserNodeData(NamedTuple):
    """A parser node."""

    role: str = None #: *str* – the role.
    pos: str = None  #: *str* – the post-tag.
    term: str = None #: *str* – the text term.

    @classmethod
    def from_text(cls, text):
        """Create a :class:`ParserNodeData` object from :class:`ckipnlp.parser.CkipParser` output."""
        fields = text.split(':')
        return cls(*fields)

    def __str__(self):
        return self.to_text()

    def to_text(self):
        """Transform to plain text."""
        return ':'.join(filter(None, self))

    def to_dict(self):
        """Transform to python dict/list."""
        return self._asdict() # pylint: disable=no-member

    def to_json(self, **kwargs):
        """Transform to JSON format."""
        return _json.dumps(self.to_dict(), **kwargs)

class ParserNode(_treelib.Node):
    """A parser node for tree.

    Attributes
    ----------
        data : :class:`ParserNodeData`

    See Also
    --------
        treelib.tree.Node: Please refer `<https://treelib.readthedocs.io/>`_ for built-in usages.
    """

    def __repr__(self):
        return '{name}(tag={tag}, identifier={identifier})'.format(
            name=self.__class__.__name__,
            tag=self.tag,
            identifier=self.identifier,
        )

    def to_text(self):
        """Transform to plain text."""
        return self.data.to_text()

    def to_dict(self):
        """Transform to python dict/list."""
        return _collections.OrderedDict(id=self.identifier, **self.data.to_dict())

    def to_json(self, **kwargs):
        """Transform to JSON format."""
        return _json.dumps(self.to_dict(), **kwargs)

class ParserRelation(NamedTuple):
    """A parser relation."""

    head: ParserNode #: :class:`ParserNode` – the head node.
    tail: ParserNode #: :class:`ParserNode` – the tail node.
    relation: str #: *str* – the relation.

    def __repr__(self):
        ret = '{name}(head={head}, tail={tail}, relation={relation})' if self.head_first \
         else '{name}(tail={tail}, head={head}, relation={relation})'
        return ret.format(name=type(self).__name__, head=self.head, tail=self.tail, relation=self.relation)

    @property
    def head_first(self):
        return self.head.identifier <= self.tail.identifier

    def to_dict(self):
        """Transform to python dict/list."""
        return _collections.OrderedDict(head=self.head.to_dict(), tail=self.head.to_dict(), relation=self.relation)

    def to_json(self, **kwargs):
        """Transform to JSON format."""
        return _json.dumps(self.to_dict(), **kwargs)

################################################################################################################################

class ParserTree(_treelib.Tree):
    """A parsed tree.

    See Also
    --------
    treereelib.tree.Tree: Please refer `<https://treelib.readthedocs.io/>`_ for built-in usages.
    """

    node_class = ParserNode

    @staticmethod
    def normalize_text(tree_text):
        """Text normalization for :class:`ckipnlp.parser.CkipParser` output.

        Remove leading number and trailing ``#``. Prepend ``root:`` at beginning.
        """
        return 'root:' + tree_text.split(' ', 2)[-1].split('#')[0]

    @classmethod
    def from_text(cls, tree_text, *, normalize=True):
        """Create a :class:`ParserTree` object from :class:`ckipnlp.parser.CkipParser` output.

        Parameters
        ----------
            text : str
                A parsed tree from :class:`ckipnlp.parser.CkipParser` output.
            normalize : str
                Do text normalization. Please refer :func:`ParserTree.normalize_text`.
        """
        if normalize:
            tree_text = cls.normalize_text(tree_text)

        tree = cls()
        node_id = 0
        node_queue = [None]
        text = ''
        ending = True

        for char in tree_text:
            if char == '(':
                node_data = ParserNodeData.from_text(text)
                tree.create_node(tag=text, identifier=node_id, parent=node_queue[-1], data=node_data)

                node_queue.append(node_id)
                node_id += 1
                text = ''

            elif char == ')':
                if not ending:
                    node_data = ParserNodeData.from_text(text)
                    tree.create_node(tag=text, identifier=node_id, parent=node_queue[-1], data=node_data)
                    node_id += 1

                node_queue.pop()
                text = ''
                ending = True

            elif char == '|':
                if not ending:
                    node_data = ParserNodeData.from_text(text)
                    tree.create_node(tag=text, identifier=node_id, parent=node_queue[-1], data=node_data)
                    node_id += 1

                text = ''
                ending = True

            else:
                ending = False
                text += char

        return tree

    def __str__(self):
        self.to_text()

    def to_text(self, node_id=0):
        """Transform to plain text."""
        node = self[node_id]
        tree_text = node.to_text()

        children_text = '|'.join((self.to_text(child.identifier) for child in self.children(node_id)))
        if children_text:
            tree_text = '{}({})'.format(tree_text, children_text)

        return tree_text

    def to_dict(self, node_id=0): # pylint: disable=arguments-differ
        """Transform to python dict/list."""
        node = self[node_id]
        tree_dict = node.to_dict()

        for child in self.children(node_id):
            tree_dict.setdefault('children', list()).append(self.to_dict(child.identifier))

        return tree_dict

    def to_json(self, **kwargs): # pylint: disable=arguments-differ
        """Transform to JSON format."""
        return _json.dumps(self.to_dict(), **kwargs)

    def show(self, *, # pylint: disable=arguments-differ
        key=lambda node: node.identifier,
        idhidden=False,
        **kwargs,
    ):
        """Show pretty tree."""
        super().show(key=key, idhidden=idhidden, **kwargs)

    def has_dummies(self, node_id):
        """Determine if a node has dummies.

        Parameters
        ----------
            node_id : int
                ID of target node.

        Returns
        -------
            bool
                whether or not target node has dummies.
        """
        roles = [node.data.role for node in self.children(node_id)]
        return 'DUMMY1' in roles and 'DUMMY2' in roles

    def get_dummies(self, node_id, deep=True, _check=True):
        """Get dummies of a node.

        Parameters
        ----------
            node_id : int
                ID of target node.
            deep : bool
                find dummies recursively.

        Returns
        -------
            Tuple[:class:`ParserNode`]
                the dummies.

        Raises
        ------
            LookupError
                when target node has no dummy (only when **_check** is set).
        """
        if _check and not self.has_dummies(node_id):
            raise LookupError('Node ({node_id}) does not have dummies!'.format(node_id=node_id))

        dummy1 = ()
        dummy2 = ()

        for child in self.children(node_id):

            if child.data.role == 'DUMMY1':
                if deep and self.has_dummies(child.identifier):
                    dummy1 = self.get_dummies(child.identifier, deep=True, _check=False)
                else:
                    dummy1 = (child,)

            if child.data.role == 'DUMMY2':
                if deep and self.has_dummies(child.identifier):
                    dummy2 = self.get_dummies(child.identifier, deep=True, _check=False)
                else:
                    dummy2 = (child,)

        return (*dummy1, *dummy2,)

    def get_heads(self, root_id=0, deep=True): # pylint: disable=too-many-branches
        """Get all head nodes of a subtree.

        Parameters
        ----------
            root_id : int
                ID of the root node of target subtree.
            deep : bool
                find heads recursively.

        Returns
        -------
            List[:class:`ParserNode`]
                the head nodes (when **deep** is set).
            :class:`ParserNode`
                the head node (when **deep** is not set).

        Todo
        ----
            Get information of nodes with pos type PP or GP.
        """
        head_nodes = None
        children = list(self.children(root_id))

        # No child, choose the root node instead
        if not children:
            head_nodes = (self[root_id],)

        # Find head
        if head_nodes is None:
            for child in children:
                if child.data.role == 'head':
                    if not deep:
                        head_nodes = (child,)
                    else:
                        if child.data.pos == 'Caa': # Found Caa, choose dummies of root instead
                            head_nodes = tuple(_itertools.chain.from_iterable(
                                self.get_heads(node.identifier) for node in self.get_dummies(root_id, _check=False)
                            ))
                        else:
                            head_nodes = self.get_heads(child.identifier)
                    break

        # Find Head
        if head_nodes is None:
            for child in children:
                if child.data.role == 'Head':
                    if not deep:
                        head_nodes = (child,)
                    else:
                        if child.data.pos == 'Caa': # Found Caa, choose dummies of root instead
                            head_nodes = tuple(_itertools.chain.from_iterable(
                                self.get_heads(node.identifier) for node in self.get_dummies(root_id, _check=False)
                            ))
                        else:
                            head_nodes = self.get_heads(child.identifier)
                    break

        # Found no head, choose the last child instead
        if head_nodes is None:
            head_nodes = (children[-1],)

        return head_nodes[0] if not deep else head_nodes

    def get_relations(self, root_id=0):
        """Get all relations of a subtree.

        Parameters
        ----------
            root_id : int
                ID of the subtree root node.

        Yields
        ------
            :class:`ParserRelation`
                the relation.
        """

        head_root_node = self.get_heads(root_id, deep=False)

        # Skip Caa
        if head_root_node.data.pos == 'Caa':
            return

        # Get heads
        for head_node in self.get_heads(root_id):

            # Get tails
            for tail in self.children(root_id):
                if tail.identifier != head_root_node.identifier:
                    if tail.data.term: # if tail is a leaf node
                        yield ParserRelation(head=head_node, tail=tail, relation=tail.data.role)
                    else:
                        for node in self.get_heads(tail.identifier):
                            yield ParserRelation(head=head_node, tail=node, relation=tail.data.role)

        # Recursion
        for child in self.children(root_id):
            yield from self.get_relations(child.identifier)
